from yta_general_utils.temp import create_temp_filename
from yta_general_utils.programming.env import get_current_project_env
from yta_general_utils.downloader import Downloader
from abc import ABC, abstractmethod
from typing import Union
from random import choice

import requests


class GifDownloader(ABC):
    """
    Abstract class to be inherited from other classes
    to download gifs.
    """

    @abstractmethod
    def download(
        self,
        keywords: str,
        output_filename: Union[str, None] = None
    ):
        """
        Download a gif found with the given 'keywords'.
        """
        pass

class GiphyGifDownloader(GifDownloader):
    """
    Gif downloader that uses the Giphy platform.
    """

    _GIPHY_API_KEY = get_current_project_env('GIPHY_API_KEY')

    def download(
        self,
        keywords: str,
        output_filename: Union[str, None] = None
    ):
        """
        Download a random gif from Giphy platform using 
        an API key.

        This method returns None if no gif found or the
        final 'output_filename' with it's been locally
        stored.

        For more information, check this:
        - https://developers.giphy.com/dashboard/
        """
        if not output_filename:
            return None
        
        limit = 5

        url = "http://api.giphy.com/v1/gifs/search"
        url += f'?q={keywords}&api_key={self._GIPHY_API_KEY}&limit={str(limit)}'

        response = requests.get(url)
        response = response.json()

        if not response or len(response['data']) == 0:
            # TODO: Raise exception of no gif found
            print(f'No gif found with the keywords "{keywords}".')
            return None
        
        if not output_filename:
            output_filename = create_temp_filename('tmp_gif.webp')

        element = choice(response['data'])
        gif_url = 'https://i.giphy.com/' + element['id'] + '.webp'

        if not output_filename.endswith('.webp'):
            output_filename += '.webp'

        Downloader.download_image(gif_url, output_filename)

        return output_filename


from dataclasses import dataclass
from typing import Union


@dataclass
class ImageResult:
    """
    Class to contain the result of any of our methods
    with images. This will return the image itself 
    and the 'output_filename' if generated.
    """

    image: any
    """
    The image once it's been modified.
    """
    output_filename: Union[str, None]
    """
    The filename in which the image has been stored
    once it's been modified, if requested. Sometimes
    you need the image stored as a file apart from
    being modified, and sometimes you don't.
    """

    def __init__(
        self,
        image: any,
        output_filename: Union[str, None],
    ):
        self.image = image
        self.output_filename = output_filename



from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from yta_general_utils.temp import create_temp_filename
from PIL import Image


class YoutubeImageDownloader:
    """
    Class to download images from the Youtube
    website.
    """

    @staticmethod
    def download_youtube_video_comments(
        url: str,
        number_of_comments: int = 5
    ):
        """
        Download (if possible) the 'number_of_comments' first
        comments of the Youtube video with the given 'url'. It
        will scrap the web page and screenshot those comments.

        The amount of comments available could be lower than 
        the 'number_of_comments' requested or even zero.

        This method will return an array containing all the
        comment images obtained, that could be empty if no one
        was available.
        """
        if not url:
            raise Exception('No "url" provided.')
        
        # TODO: Check if url is a valid Youtube url

        if not number_of_comments:
            number_of_comments = 5

        scrapper = ChromeScrapper(False)
        # Example of url: 'https://www.youtube.com/watch?v=OvUj2WsADjI'
        scrapper.go_to_web_and_wait_util_loaded(url)
        # We need to scroll down to let the comments load
        # TODO: This can be better, think about a more specific strategy
        # about scrolling
        scrapper.scroll_down(1000)
        scrapper.wait(1)
        scrapper.scroll_down(1000)
        scrapper.wait(1)

        # We need to make sure the comments are load
        scrapper.find_element_by_element_type_waiting('ytd-comment-thread-renderer')
        comments = scrapper.find_elements_by_element_type('ytd-comment-thread-renderer')

        if len(comments) >= number_of_comments:
            comments[:5]

        # We remove the header bar to avoid being over our comments in some cases
        youtube_top_bar = scrapper.find_element_by_id_waiting('masthead-container')
        scrapper.remove_element(youtube_top_bar)

        screenshots = []
        for comment in comments:
            # TODO: I need to close the 'No, gracias' 'Probar 1 mes' popup
            scrapper.scroll_to_element(comment)
            style = 'width: 500px; padding: 10px;'
            scrapper.set_element_style(comment, style)
            filename = create_temp_filename('tmp_comment_screenshot.png')
            scrapper.screenshot_element(comment, filename)
            screenshots.append(Image.open(filename))

        # TODO: Maybe I need to return the filenames instead (?)
        return screenshots


"""
When I get the 'innerText' attribute from a comment,
I receive it with a specific structure:

print(comment.get_attribute('innerText'))
# I can handle the information from this innertext
# If I split by \n:

# 1st is author (@pabloarielcorderovillacort2149)
# 2nd is date (hace 6 meses)
# 3rd and next ones are comment text
# Penultimate is the number of likes (number)
# Last one is 'Responder'
#
# This below is an example of a read comment:
#
# @pabloarielcorderovillacort2149
# hace 6 meses
# No puedo esperar tu siguiente vídeo.

# Andy Serkis es una joya de actor, Gollum y Cesar son mis personajes favoritos de este actor.
# 3
# Responder
"""


from yta_image.parser import ImageParser
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.downloader import Downloader
from yta_general_utils.programming.env import get_current_project_env
from typing import Union

import time
import requests


PRODIA_API_KEY =  get_current_project_env('PRODIA_API_KEY')

def generate_image_with_prodia(
    prompt: str,
    output_filename: Union[str, None] = None
):
    # TODO: Raise exception
    if not prompt:
        return None
    
    # If you comment this and uncomment the one below it works
    # seed = randint(1000000000, 9999999999)
    # response = requests.get('https://api.prodia.com/generate?new=true&prompt=' + prompt + '&model=absolutereality_v181.safetensors+%5B3d9d4d2b%5D&steps=20&cfg=7&seed=' + str(seed) + '&sampler=DPM%2B%2B+2M+Karras&aspect_ratio=square')
    payload = {
        'new': True,
        'prompt': prompt,
        #'model': 'absolutereality_v181.safetensors [3d9d4d2b]',   # this model works on above request, not here
        'model': 'sd_xl_base_1.0.safetensors [be9edd61]',
        #'negative_prompt': '',
        'steps': 20,
        'cfg_scale': 7,
        'seed': 2328045384,
        'sampler': 'DPM++ 2M Karras',
        'width': 1344,
        'height': 768
    }
    headers = {
        "accept": "application/json",
        "content-type": "application/json",
        "X-Prodia-Key": PRODIA_API_KEY
    }
    url = 'https://api.prodia.com/v1/sdxl/generate'
    response = requests.post(url, json = payload, headers = headers)
    response = response.json()

    # TODO: Improve 'output_filename' handling
    if output_filename is None:
        output_filename = create_temp_filename('tmp_prodia.png')

    # When requested it is queued, so we ask for it until it is done
    if "status" in response and response['status'] == 'queued':
        job_id = response['job']
        image_filename = retrieve_job(job_id, output_filename)
    else:
        print(response)
        raise Exception('There was an error when generating a Prodia AI Image.')
    
    image = ImageParser.to_pillow(image_filename) if image is not None else None

    return image, output_filename

def retrieve_job(
        job_id: str,
        output_filename: Union[str, None] = None
    ):
        """
        Makes a request for the image that is being
        generated with the provided 'job_id'.

        It has a loop to wait until it is done. This
        code is critic because of the loop.
        """
        url = f'https://api.prodia.com/v1/job/{str(job_id)}'

        headers = {
            'accept': 'application/json',
            'X-Prodia-Key': PRODIA_API_KEY
        }

        response = requests.get(url, headers = headers)
        response = response.json()
        #print(response)

        # TODO: Do a passive waiting
        is_downloadable = True

        if response['status'] != 'succeeded':
            is_downloadable = False

        # TODO: Implement a tries number
        while not is_downloadable:
            time.sleep(5)
            print('Doing a request in loop')

            # We do the call again
            response = requests.get(url, headers = headers)
            response = response.json()
            print(response)
            if 'imageUrl' in response:
                is_downloadable = True

        image_url = response['imageUrl']

        return Downloader.download_image(image_url, output_filename)


from yta_image.parser import ImageParser
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.downloader import Downloader
from yta_general_utils.url import encode_url_parameter
from typing import Union


def generate_image_with_pollinations(
    prompt: str,
    output_filename: Union[str, None] = None
):
    # TODO: Improve 'output_filename' handling
    if output_filename is None:
        output_filename = create_temp_filename('ai_pollinations.png')
    
    prompt = encode_url_parameter(prompt)

    # TODO: Make some of these customizable
    WIDTH = 1920
    HEIGHT = 1080
    # TODO: This seed should be a random value or
    # I will receive the same image with the same
    # prompt
    SEED = 43
    MODEL = 'flux'

    url = f'https://pollinations.ai/p/{prompt}?width={WIDTH}&height={HEIGHT}&seed={SEED}&model={MODEL}'

    image_filename = Downloader.download_image(url, output_filename)
    
    image = ImageParser.to_pillow(image_filename) if image is not None else None

    return image, output_filename

"""
    Check because there is also a model available for
    download and to work with it (as they say here
    https://pollinations.ai/):

    # Using the pollinations pypi package
    ## pip install pollinations

    import pollinations as ai

    model_obj = ai.Model()

    image = model_obj.generate(
        prompt=f'Awesome and hyperrealistic photography of a vietnamese woman... {ai.realistic}',
        model=ai.flux,
        width=1038,
        height=845,
        seed=43
    )
    image.save('image-output.jpg')

    print(image.url)
    """


from yta_image.parser import ImageParser
from yta_general_utils.programming.output import handle_output_filename
from yta_general_utils.file.enums import ImageFileExtension
from PIL import ImageDraw, Image
from typing import Union
from abc import ABC, abstractmethod

import numpy as np


class ImageMaskPreset(ABC):
    """
    Abstract class to be implemented by the custom
    image masks.
    """
    
    @abstractmethod
    def apply_on(
        image: Union[str, Image.Image, np.ndarray],
        output_filename: Union[str, None] = None
    ):
        pass

class RoundedCornersImageMaskPreset(ImageMaskPreset):
    def apply_on(
        image: Union[str, Image.Image, np.ndarray],
        output_filename: Union[str, None] = None
    ):
        """
        Generate a mask of the provided 'image' with the corners
        rounded and return it as a not normalized RGB Pillow Image 
        with all values between 0 (white) and 255 (black). This 
        means that the result, if converted to array, will be
        [0, 0, 0] for each white pixel and [255, 255, 255] for 
        black ones.

        Thank you: https://github.com/Zulko/moviepy/issues/2120#issue-2141195159
        """
        image = ImageParser.to_pillow(image)

        # Create a whole black image of the same size
        mask = Image.new('L', image.size, 'black')
        mask_drawing = ImageDraw.Draw(mask)

        # Generate the rounded corners mask
        w, h = image.size
        radius = 20
        # Rectangles to cover
        mask_drawing.rectangle([radius, 0, w - radius, h], fill = 'white')
        mask_drawing.rectangle([0, radius, w, h - radius], fill = 'white')
        # Circles at the corners: TL, TR, BL, BR
        mask_drawing.ellipse([0, 0, 2 * radius, 2 * radius], fill = 'white')
        mask_drawing.ellipse([w - 2 * radius, 0, w, 2 * radius], fill = 'white')
        mask_drawing.ellipse([0, h - 2 * radius, 2 * radius, h], fill = 'white')
        mask_drawing.ellipse([w - 2 * radius, h - 2 * radius, w, h], fill = 'white')

        if output_filename is not None:
            output_filename = handle_output_filename(output_filename, ImageFileExtension.PNG)
            mask.save(output_filename)

        return mask
    
class TicketImageMaskPreset(ImageMaskPreset):
    def apply_on(
        image: Union[str, Image.Image, np.ndarray],
        output_filename: Union[str, None] = None
    ):
        """
        Create triangles on top and bottom to simulate
        a shop printed ticket over the given 'image'.
        """
        image = ImageParser.to_pillow(image)

        # Create a whole black image of the same size
        mask = Image.new('L', image.size, 'white')
        mask_drawing = ImageDraw.Draw(mask)

        # Generate the mask
        w, h = image.size
        triangle_size = 15

        for i in range(0, w, triangle_size):
            mask_drawing.polygon([(i, h), (i + triangle_size, h), (i + triangle_size // 2, h - triangle_size)], fill = 'black')
            mask_drawing.polygon([(i, 0), (i + triangle_size, 0), (i + triangle_size // 2, triangle_size)], fill = 'black')

        if output_filename is not None:
            output_filename = handle_output_filename(output_filename, ImageFileExtension.PNG)
            mask.save(output_filename)

        return mask




from yta_image.parser import ImageParser
from PIL import Image


class ImageValidator:
    """
    Class to simplify and encapsulate the image
    validation.
    """

    @staticmethod
    def is_file_valid_image(image_filename: str):
        """
        Tries to open the 'image_filename' provided to
        check if it is corrupt or it is valid. It returns
        True if the provided image is valid, or False if
        is corrupt.
        """
        try:
            im = Image.open(image_filename)
            im.verify()
            im.close()
        except (IOError, OSError, Image.DecompressionBombError) as e:
            return False
            
        return True

def has_transparency(image: Image):
    """
    Checks if the provided image (read with pillow) has transparency.
    """
    image = ImageParser.to_pillow(image)

    # TODO: What about 'image' filename instead of Image PIL (?)
    if image.info.get("transparency", None) is not None:
        return True
    if image.mode == "P":
        transparent = image.info.get("transparency", -1)
        for _, index in image.getcolors():
            if index == transparent:
                return True
    elif image.mode == "RGBA":
        extrema = image.getextrema()
        if extrema[3][0] < 255:
            return True

    return False


"""
AI Image generation file that contains the classes
capable to generate AI images. These classes will
raise Exceptions if the parameters are not provided
and will use internal functionality to do it.

Programmer help: The classes implement parameter
validation and raise exceptions, but the other files
from which other methods are imported do not 
implement them, so make sure you pass the right
and expected parameters. This should be the ideal
structure to keep in code, but you know... I write
code very fast so I can't go back and review and
refactor code often... So sorry about that :P
"""
from yta_image.generation.ai.prodia import generate_image_with_prodia
from yta_image.generation.ai.pollinations import generate_image_with_pollinations
from yta_general_utils.programming.parameter_validator import PythonValidator
from abc import ABC, abstractmethod
from typing import Union


class AIImageGenerator(ABC):
    """
    Abstract class to be inherited by any specific
    AI image generator.
    """

    @abstractmethod
    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        """
        Generate an image with the given 'prompt' and
        store it locally if 'output_filename' is 
        provided.
        """
        pass

class ProdiaAIImageGenerator(AIImageGenerator):
    """
    Prodia AI image generator.
    """

    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        """
        Generate an AI image and return 2 values, the
        image read with pillow and the final output
        filename used to store the image locally.
        """
        # TODO: Validate params and raise exception
        image, output_filename = generate_image_with_prodia(prompt, output_filename)

        return image, output_filename
    
class PollinationsAIImageGenerator(AIImageGenerator):
    """
    Pollinations AI image generator.

    This is using the Pollinations platform wich
    contains an AI image generator API and 
    open-source model.

    Source: https://pollinations.ai/
    """

    def generate_image(
        self,
        prompt: str,
        output_filename: Union[str, None] = None
    ):
        """
        Generate an image with the Pollinations AI image generation model
        using the provided 'prompt' and stores it locally as 
        'output_filename'.
        """
        # TODO: Validate params and raise exception
        if not PythonValidator.is_string(prompt):
            raise Exception('Provided "prompt" parameter is not a valid prompt.')
        
        image, output_filename = generate_image_with_pollinations(prompt, output_filename)

        return image, output_filename

    


"""
This is a module in which I keep some code
that could be not preserved in next versions.
"""
from yta_image.parser import ImageParser
from PIL import Image
from typing import Union


# TODO: Remove this method in a few commits if still
# not used
def image_has_transparency(image: Union[Image.Image, str]):
    """
    Checks if the provided image (read with pillow) has transparency.
    This method returns True if yes or False if not.
    """
    image = ImageParser.to_pillow(image)

    if image.info.get("transparency", None) is not None:
        return True
    if image.mode == "P":
        transparent = image.info.get("transparency", -1)
        for _, index in image.getcolors():
            if index == transparent:
                return True
    elif image.mode == "RGBA":
        extrema = image.getextrema()
        if extrema[3][0] < 255:
            return True

    return False


from yta_image.mask.presets import ImageMaskPreset
from yta_image.parser import ImageParser
from yta_general_utils.programming.parameter_validator import PythonValidator
from PIL import Image
from typing import Union

import numpy as np
import cv2


class ImageMask:
    """
    Class to encapsulate and simplify the functionality
    related to image masking.
    """

    @staticmethod
    def apply_mask_preset(
        image: Union[str, Image.Image, np.ndarray],
        mask_preset: ImageMaskPreset,
        output_filename: Union[str, None] = None
    ):
        """
        Apply the provided 'mask' to the given 'image'.

        This method return the result image as a ndim = 4 numpy
        array of not normalized values between 0 and 255, even
        the alpha channel.
        """
        if not PythonValidator.is_subclass(mask_preset, ImageMaskPreset):
            raise Exception('The provided "mask_preset" parameter is not a valid ImageMaskPreset class instance.')

        image = ImageParser.to_numpy(image, mode = 'RGB')

        # Obtain mask as ndim = 1 where black is transparent
        mask = ImageParser.to_numpy(mask_preset.apply_on(image))[:, :, 1]
        
        # Add the mask to the Pillow Image
        image = np.dstack((image, mask))

        if output_filename is not None:
            # TODO: Apply 'output' handler
            cv2.imwrite(output_filename, image)

        # TODO: Maybe convert to pillow?

        return image